
#include <set>
#include <iostream>
using namespace std;

#include <stdlib.h>
#include <time.h>

namespace Mundo
{
class Inseto
{
    private:
        static int cont;
        int energia;

    private:
        virtual void variar() = 0;

    protected:
        void informar_energia() const { cout << "O valor da energia eh: " << energia << endl; }

    public:

        Inseto():energia (cont++) { }

        virtual ~Inseto() { energia = -1; }

        //void setEnergia(const int e) { energia = e; }
        const int getEnergia() const { return energia;}
        void operator--() { if (energia) energia--; }

        const bool operator>(Inseto& I) const { return (getEnergia() > I.getEnergia()); }

        virtual void executar() = 0;
};
int Inseto::cont(1);

class Colonia
{
    private:
        set<Inseto*> conj;
    public:

        Colonia(): conj() { conj.clear(); }
        ~Colonia()
        {
            //set<Inseto*>::iterator it;
            set<Inseto*>::const_iterator it;
            it = conj.begin();

            while (it != conj.end())
            {
                if (*it)
                {
                    //conj.erase(it);
                    delete *it;
                }
                it++;
            }
            conj.clear();
        }

        void incluir(Inseto* pi) { if (pi) { conj.insert(pi); } else { cout << "Ponteiro nulo!" << endl; } }

        void executar()
        {
            set<Inseto*>::const_iterator iterador;

            for ( iterador = conj.begin(); iterador != conj.end(); iterador++)
            {
                if (*iterador)
                {
                    cout << endl; (*iterador)->executar();
                }
                else
                {
                    cout << "Ponteiro nulo no conjunto" << endl;
                }
            }

            cout << endl;
        }
};

class Abelha : public Inseto
{
    private:
        bool ferrao;

    private:
        void variar() { ferrao = !ferrao; informar_energia(); }

    public:

        Abelha(): Inseto(), ferrao((bool)rand()%2)
        {
            time_t t;                   // https://pt.wikipedia.org/wiki/Time_t
            srand((unsigned) time(&t)); // Para funo srand e rand vide: https://www.tutorialspoint.com/c_standard_library/c_function_srand.htm
        }

        ~Abelha() { ferrao = false; }

        void setFerrao(const bool f) { ferrao = f; } // at o implementei mas no  usado finalmente...
        const int getFerrao() const { return ferrao; }

        void executar() { variar(); }
};

class Formiga : public Inseto
{
    private:
        Abelha* pAbelha;

    private:
        void interagir()
        {
            if ( pAbelha )
            {
                if ( ! pAbelha->getFerrao()  )
                {
                    //if ( getEnergia() > pAbelha->getEnergia()) // ou
                    //if ( operator>(*pAbelha) // ou
                    //if ( (*this ) > (*pAbelha) ) // ou
                    if ( operator> ( *( static_cast<Inseto*>(pAbelha) ) ) )
                    {
                        pAbelha->operator--();
                        cout << "Ataquei!" << endl;
                    }
                    else
                    {
                        cout << "No ataquei pois estou com menos energia!" << endl;
                    }
                }
                else
                {
                    cout << "No ataquei pois a abelha  ferroenta!" << endl;
                }
            }
            else
            {
                cout << "Ponteiro de abelho nulo!" << endl;
            }
        }

        void variar() { if (!(rand()%10)) { operator--(); }  informar_energia(); }

    public:

        Formiga(): Inseto(), pAbelha(NULL)
        {
            time_t t;                   // https://pt.wikipedia.org/wiki/Time_t
            srand((unsigned) time(&t)); // Para funo srand e rand vide: https://www.tutorialspoint.com/c_standard_library/c_function_srand.htm
        }

        ~Formiga() { pAbelha = NULL; }

        void setAbelha(Abelha* pAb) { pAbelha = pAb; }

        Abelha* getAbelha() const { return pAbelha; }

        void executar() { interagir(); variar(); }
};

class Habitat
{
    private:
        Colonia* pColonia;

    private:

        void criarInsetos()
        {
            Abelha* pAbelha = NULL;
            Formiga* pFormiga = NULL;
            const int max = 20;

            for (int i = 0; i < max; i++)
            {
                pAbelha = new Abelha();
                if (pAbelha)
                {
                    pColonia->incluir(static_cast<Inseto*>(pAbelha));

                    pFormiga = new Formiga();
                    if (pFormiga)
                    {
                        pFormiga->setAbelha(pAbelha);
                        if (pColonia)
                        {
                            pColonia->incluir(static_cast<Inseto*>(pFormiga));
                        }
                        else
                        {
                            cout << "pColonia nulo!" << endl;
                        }
                    }
                    else
                    {
                        cout << "Formiga no alocada!" << endl;
                    }
                }
                else
                {
                    cout << "Abelha no alocada!" << endl;
                }
                pAbelha = NULL;
                pFormiga = NULL;
            }
        }

    public:

        Habitat(): pColonia(NULL)
        {
            time_t t;                   // https://pt.wikipedia.org/wiki/Time_t
            srand((unsigned) time(&t)); // Para funo srand e rand vide: https://www.tutorialspoint.com/c_standard_library/c_function_srand.htm

            pColonia = new Colonia();

            if (pColonia)
            {
                criarInsetos();
                executar();
            }
            else
            {
                cout << "Objeto de pColonia no intanciado dinamicamente!" << endl;
            }
        }
        ~Habitat()
        {
            if (pColonia)
            {
                delete pColonia;
                pColonia = NULL;
            }
        }

        void executar()
        {
            if (pColonia)
            {
                pColonia->executar();
            }
            else
            {
                cout << "pColonia nulo!" << endl;
            }
        }
};
} using namespace Mundo;

int main()
{

    Habitat habitat;

    cout << endl;
    cout << "FIM" << endl;

    return 0;
}

